/*
Johannes Philippi Dez.2006

Very basic CAN functionality demo program. Works fine with AVR Studio 4.12

Have fun!
*/

#include <avr/io.h>           		//Zugriff auf IO's des Controllers
#include <avr/signal.h>
#include <inttypes.h>


/*******************************************************************
definitions
*******************************************************************/
#define NO_MOBS 15
#define NOMOB   0xff

#define setbit(ADR,BIT) (ADR|=(1<<BIT))
#define clearbit(ADR,BIT) (ADR&=~(1<<BIT))


// CAN2.0 B, extendef frame(29bit) message
typedef struct
	{	
		uint32_t id;                  
		uint8_t length;   
		uint8_t data[8];  
	}CAN_message;


/*******************************************************************
Get idex of MOb for which an interrupt was detected.
*******************************************************************/
uint8_t getmob( uint32_t bitmask) 
{
 uint8_t mob;

 if( bitmask==0) return NOMOB;

 //the bitmask is shifted to the rigth until the LSB is one
 //the number of such shift operations is the index of the MOb
 for( mob=0; (bitmask & 0x01)==0; bitmask >>= 1, ++mob);

 //security check
 if ( mob > 14 ) return NOMOB;
 	else return mob;
}     

/*******************************************************************
CAN Receive Interrupt Service Routine
*******************************************************************/
SIGNAL( SIG_CAN_INTERRUPT1)
{
	uint8_t i;
 	static CAN_message message;		
 	uint8_t save_canpage = CANPAGE;		// store current CANPAGE

    //get indx of MOb for which the INT occured
	uint8_t mob = getmob( CANSIT2 | (CANSIT1 << 8));
    
	if( mob == NOMOB )   return;
    
	//select concerned MOb page
	CANPAGE = mob << 4;	
    
	//get data of selected MOb
	for (i=0; i<8; i++)		
		message.data[i] = CANMSG;
	
	//get ID of selected MOb
	message.id = 0;
	message.id |= ((uint32_t) CANIDT1<<24);	//get identifier
	message.id |= ((uint32_t) CANIDT2<<16);
	message.id |= ((uint32_t) CANIDT3<<8);
	message.id |= (CANIDT4&0xF8);
	message.id >>= 3;
	
	//do something with the received message. store data, write data, calculate something, set flags...
	/*
	-
	-
	-
	*/
    
	// reset INT reason
	CANSTMOB &= ~(1<<RXOK);					
    // re-enable RX, reconfigure MOb IDE=1
	CANCDMOB = (1<<CONMOB1) | (1<<IDE);    	
    
    CANPAGE=save_canpage;   	// restore CANPAGE
}


/*******************************************************************
CAN initialization
*******************************************************************/
void can_init( void)
{
  uint8_t mob;

	//set Baud Rate Prescaler for CAN controller bit timing (page236, table on page261)
  CANBT1 = 0x06 ;
 	//set propagation time segment (page232, table on page261)
  CANBT2 = 0x04;
 	//set PHaseSegment2 and PHaseSegment1 (page233, table on page261)
  CANBT3 = 0x13;
       
  	//for all Message Objects: set the StatusRegister CANSTMOB to zero; Disable MOb CANCDMOB=0, (CONMOB1:0=00)
  for (mob = 0; mob < NO_MOBS; mob++)
  {
     CANPAGE  = (mob << 4);
     CANSTMOB = 0;
     CANCDMOB = 0;
  }
  
  	//Set CAN-driver in enable mode (not standby-mode)
  CANGCON |= (1<<ENASTB);
  
  	//Enable CAN Interrupts globally, set enable CANRX interrupt
  CANGIE=(1<<ENIT) | (1<< ENRX);
  
  	//Set MOB for which Interrupts are enabled
  CANIE2 = (1<<IEMOB0) |  (1<<IEMOB1) |  (1<<IEMOB2) |  (1<<IEMOB3); 		
  CANIE1 = 0;

  	//reset the Status Interrupt registers
  CANSIT2 = 0x00;
  CANSIT1 = 0x00;

  	//wait until module is ready EnableFlag is on.
  while (!(CANGSTA & (1<<ENFG)));


}//can_init

/*******************************************************************
CAN transmission via mailbox 
*******************************************************************/
void can_tx (CAN_message msg)
{	
uint8_t i, mob;	

	//select MOb to be sent
	mob=1;

	//enable MOb mob, auto increment index, start with index = 0
	CANPAGE = (mob<<4);
	
	//set IDE bit, length = 8
	CANCDMOB = (1<<IDE) | (8<<DLC0);			
	
	//write 29 Bit identifier
	msg.id <<= 3;								
	CANIDT4 = (unsigned char) (msg.id&0xF8);
	CANIDT3 = (unsigned char) (msg.id>>8);
	CANIDT2 = (unsigned char) (msg.id>>16);
	CANIDT1 = (unsigned char) (msg.id>>24);
	
	//put data in mailbox
	for (i=0; i<8; i++)							
		CANMSG = msg.data [i];
	
	//enable transmission
	CANCDMOB |= (1<<CONMOB0);					
	
	//wait until sending message complete
	while (!(CANSTMOB & (1<<TXOK)));			

	//reset flag
	CANSTMOB &= ~(1<<TXOK);						

}//can_tx


/*******************************************************************
MAIN PROGRAM ENTRY POINT
*******************************************************************/
int main (void) 
{
CAN_message msg0,msg1,msg2,msg3;
uint8_t mob;

	DDRC = 0xff;					//PC0...PC7 als Ausgnge definieren fr LEDs

	can_init();						//initialize CAN driver
	sei();							//globally enable interrupts

	//select MOb to be configured
	mob = 1;
	//configure a MessageObject
	CANPAGE = (mob<<4);				//select MOb page	
	
	msg0.id = 0x0123;				//define an ID for the message
	CANIDT4 |= msg0.id << 3;		
	CANIDT3 = msg0.id >> 5;	
	CANIDT2 = msg0.id >> 13;	
	CANIDT1 = msg0.id >> 21;	
	
	// define an ID mask for the MOb. [1 = check bit; 0 = ignore bit]
	CANIDM4 = (1<<IDEMSK);			//set IDEMSK to 1 -> extended identifier
	CANIDM3 = 0x00;
	CANIDM2 = 0x00;
	CANIDM1 = 0x00;

	//configure message as receive-msg (see CANCDMOB register, page257)
	CANCDMOB |= (1<<CONMOB1) | (1<<IDE);
	 
  	//set data for message to be sent
	msg2.id = 0x0002;
	
	msg2.data[0] = 0xAA;
	msg2.data[1] = 0xBB;
	msg2.data[2] = 0xCC;
	msg2.data[3] = 0xAA;
	msg2.data[4] = 0xBB;
	msg2.data[5] = 0xCC;
	msg2.data[6] = 0xAA;
	msg2.data[7] = 0xBB;

	
	for (;;) 						// inf. program loop
	{						
		setbit(PORTC,PC0);			// PC0,PC1 on, PC2,PC3 off
		setbit(PORTC,PC1);
		clearbit(PORTC,PC2);
		clearbit(PORTC,PC3);
		
		can_tx(msg2);				//transmit one message.
			
		clearbit(PORTC,PC0);		// PC0,PC1 off, PC2,PC3 on
		clearbit(PORTC,PC1);
		setbit(PORTC,PC2);
		setbit(PORTC,PC3);
		
	} //for
}//main
